==> serializer-0.8.1.tar.gz: lib/src/core/api.dart <==
// Copyright (c) 2016, the Serializer project authors.  Please see the AUTHORS file

// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'dart:convert';

import '../codecs/type_codec.dart';

bool isPrimaryType(Type obj) => obj == num || obj == String || obj == bool || obj == int || obj == double;

bool isPrimaryTypeString(String obj) =>
    obj == "num" || obj == "String" || obj == "bool" || obj == "int" || obj == "double";

final String MapTypeString = {}.runtimeType.toString();
final String ListTypeString = [].runtimeType.toString();

Map cleanNullInMap(Map<String, dynamic> map) {
  Iterable<String> keys = new List.from(map.keys);
  keys.forEach((String k) {
    if (map[k] == null) {
      map.remove(k);
    }
  });
  return map;
}

abstract class _Serializer {
  /// Return typeInfoKey value.
  String get typeInfoKey;

  /// Return useTypeInfo value.
  bool get useTypeInfo;

  /// Return if typeInfoKey must be enabled
  bool enableTypeInfo(bool useTypeInfo, bool withTypeInfo) =>
      useTypeInfo != null ? useTypeInfo || (withTypeInfo ?? false) : this.useTypeInfo || (withTypeInfo ?? false);

  /// Registers a [typeCodec] for the specific [type].
  addTypeCodec(Type type, TypeCodec typeCodec);

  /// Registers a map of [typeCodec].
  addAllTypeCodecs(Map<String, TypeCodec> typesCodecs);

  /// Checks if a TypeCodec is registered for the [type].
  bool hasTypeCodec(Type type);

  /// Get the TypeCodec for the specific [type]
  TypeCodec typeCodec(Type type);

  /// Checks if a class is registered as a Serializable class.
  bool isSerializable(Type type);

  /// Convert the object to a Map
  Map<String, dynamic> toMap(Object input, {bool useTypeInfo, bool withTypeInfo});

  /// Convert to a Map or a List recursively
  Object toPrimaryObject(Object input, {bool useTypeInfo, bool withTypeInfo});

  /// Encode the object to serialized string
  String encode(Object input, {bool useTypeInfo, bool withTypeInfo});

  /// Decode the object from a seriablized string
  Object decode(dynamic encoded, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo});

  /// Convert a serialized object to map
  Object fromMap(Map<String, dynamic> map, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo});

  /// Convert a serialized object's [list] to a list of the given [type]
  List fromList(List<dynamic> list, {Type type, bool useTypeInfo, bool withTypeInfo});
}

/// Utility class to access to the serializer api
class Serializer extends _Serializer {
  ///////////////////
  // Public
  /////////////////////////////////////////////////////////////////////////////

  /// Dump serializable classes
  //static dumpSerializables() => printAndDumpSerializables();

  final Codec _codec;
  final String _typeInfoKey;
  final bool _useTypeInfo;
  final Map<String, TypeCodec> _typeCodecs = <String, TypeCodec>{};

  String get typeInfoKey => _typeInfoKey;
  bool get useTypeInfo => _useTypeInfo;

  Serializer({Map<String, TypeCodec> typesCodecs, Codec codec: json, String typeInfoKey: "@type", useTypeInfo: false})
      : _codec = codec,
        _typeInfoKey = typeInfoKey,
        _useTypeInfo = useTypeInfo {
    addAllTypeCodecs(typesCodecs);
  }

  factory Serializer.json({Map<String, TypeCodec> typesCodecs}) {
    return new Serializer(typesCodecs: typesCodecs, codec: json);
  }

  factory Serializer.typedJson({Map<String, TypeCodec> typesCodecs}) {
    return new Serializer(typesCodecs: typesCodecs, codec: json, useTypeInfo: true);
  }

  addTypeCodec(Type type, TypeCodec typeCodec) => _typeCodecs[type.toString()] = typeCodec;

  @override
  addAllTypeCodecs(Map<String, TypeCodec> typesCodecs) {
    _typeCodecs.addAll(typesCodecs ?? {});
  }

  bool hasTypeCodec(Type type) => _typeCodecs.values?.any((TypeCodec tc) => tc.type == type);

  TypeCodec typeCodec(Type type) =>
      _typeCodecs.values.firstWhere((TypeCodec tc) => tc.type == type, orElse: () => null);

  TypeCodec typeCodecFromString(String type) =>
      _typeCodecs.values.firstWhere((TypeCodec tc) => tc.typeInfo == type, orElse: () => null);

  bool isSerializable(Type type) => hasTypeCodec(type);

  Map<String, dynamic> toMap(Object input, {bool useTypeInfo, bool withTypeInfo}) =>
      toPrimaryObject(input, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo) as Map<String, dynamic>;

  /// Convert to a Map or a List recursively
  Object toPrimaryObject(Object input, {bool useTypeInfo, bool withTypeInfo}) =>
      _encodeValue(input, withReferenceable: true, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);

  /// Encode the object to serialized string
  String encode(Object input, {bool useTypeInfo, bool withTypeInfo}) =>
      _encode(input, withReferenceable: true, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);

  /// Decode the object from a seriablized string
  Object decode(dynamic encoded, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo}) =>
      _decode(encoded, type: type, mapOf: mapOf, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);

  /// Convert a serialized object to map
  Object fromMap(Map map, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo}) =>
      _fromMap(map, type: type, mapOf: mapOf, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);

  /// Convert a serialized object's [list] to a list of the given [type]
  List<dynamic> fromList(List list, {Type type, bool useTypeInfo, bool withTypeInfo}) =>
      _fromList(list, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);

  ///////////////////
  // Private
  /////////////////////////////////////////////////////////////////////////////
  List<Type> _findGenericOfMap(Type type) {
    String str = type.toString();
    RegExp reg = new RegExp(r"^Map<(.*)\ *,\ *(.*)>$");
    Iterable<Match> matches = reg.allMatches(str);
    if (matches == null || matches.isEmpty) {
      return null;
    }
    return [_decodeType(matches.first.group(1)), _decodeType(matches.first.group(2))];
  }

  Type _findGenericOfList(Type type) {
    String str = type.toString();
    RegExp reg = new RegExp(r"^List<(.*)>$");
    Iterable<Match> matches = reg.allMatches(str);
    if (matches == null || matches.isEmpty) {
      return null;
    }
    return _decodeType(matches.first.group(1));
  }

  String _getCorrectType(String name) {
    /*   String type = correspondingMinifiedTypes[name]; //todo: check minified
    if (type != null && _classes.containsKey(type)) {
      return type;
    }*/
    return name;
  }

  Type _decodeType(String name) {
    name = _getCorrectType(name);
    switch (name) {
      case "num":
        return num;
      case "String":
        return String;
      case "bool":
        return bool;
      case "int":
        return int;
      case "double":
        return double;
      case "Map":
      case "_JsonMap":
        return Map;
      case "List":
        return List;
      default:
        if (name == MapTypeString) {
          return Map;
        } else if (name == ListTypeString) {
          return List;
        }
        return typeCodecFromString(name)?.type;
    }
  }

  Object _decode(dynamic encoded, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo}) {
    if (encoded == null) {
      return null;
    }
    dynamic value = encoded;
    if (value is String) {
      if (value.isEmpty) {
        return null;
      }
      if ((value.startsWith("{") && value.endsWith("}")) || (value.startsWith("[") && value.endsWith("]"))) {
        value = _codec.decode(encoded);
      } else {
        value = num.tryParse(encoded) ?? value;
      }
    }
    if (value is Map) {
      return _fromMap(value, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo, mapOf: mapOf);
    } else if (value is List) {
      return _fromList(value, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    }
    return _decodeValue(value, type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
  }

  Object _fromMap(Map map, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo}) {
    if (map == null || map.isEmpty) {
      return null;
    }

    if ((enableTypeInfo(useTypeInfo, withTypeInfo) || type == dynamic) && map.containsKey(_typeInfoKey)) {
      type = _decodeType(map[_typeInfoKey]);
    } else {
      type ??= Map;
    }

    // Only Map
    if (type == Map) {
      Type embedType = mapOf != null && mapOf.length >= 2 ? mapOf[1] : null;
      Map data = new Map();
      map.forEach((key, value) => data[key.toString()] =
          embedType == null && useTypeInfo == null ? value : _decodeValue(value, embedType, useTypeInfo: useTypeInfo));
      return data;
    }
    return _decodeValue(map, type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
  }

  List<dynamic> _fromList(List list, {Type type, bool useTypeInfo, bool withTypeInfo}) {
    List<dynamic> data = new List<dynamic>();
    list.forEach((value) => data.add(_decodeValue(value, type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo)));
    return data;
  }

  Object _decodeValue(Object value, Type type, {bool useTypeInfo, bool withTypeInfo}) {
    if (value == null) {
      return null;
    }
    String t = value.runtimeType.toString();
    if (value is Map && (_useTypeInfo == true || useTypeInfo == true) && value[typeInfoKey] != null) {
      t = value[typeInfoKey];
    }
    type ??= _decodeType(t);
    if (isPrimaryType(type) ||
        value.runtimeType == type ||
        (isPrimaryType(value.runtimeType) && hasTypeCodec(type) == false)) {
      return value;
    } else if (hasTypeCodec(type) == true) {
      TypeCodec codec = typeCodec(type);
      dynamic val = codec.decode(value, serializer: this);
      return val;
    } else if (type.toString().startsWith("Map") || type.toString().startsWith("List")) {
      if (value is Map) {
        if (type.toString().startsWith("Map")) {
          return _fromMap(value,
              type: Map, mapOf: _findGenericOfMap(type), useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
        }
        return _fromMap(value, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
      } else if (value is List) {
        if (type.toString().startsWith("List") && value is List) {
          return _fromList(value, type: _findGenericOfList(type), useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
        }
        return _fromList(value, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
      }
    }

    return null;
  }

  Object _encodeValue(Object value, {bool withReferenceable: false, Type type, bool useTypeInfo, bool withTypeInfo}) {
    if (isPrimaryType(value.runtimeType) == true) {
      return value;
    } else if (hasTypeCodec(value.runtimeType) == true) {
      TypeCodec codec = typeCodec(value.runtimeType);
      return codec.encode(value, serializer: this, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    } else if (value is Map) {
      return _encodeMap(value,
          withReferenceable: withReferenceable, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    } else if (value is List) {
      return _encodeList(value,
          withReferenceable: withReferenceable, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    }
    return null;
  }

  Object _encodeMap(Map value, {bool withReferenceable: false, Type type, bool useTypeInfo, bool withTypeInfo}) {
    Map<String, dynamic> map = new Map<String, dynamic>();

    value.forEach((dynamic key, dynamic val) {
      map[key.toString()] = toPrimaryObject(val, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    });

    return map;
  }

  List _encodeList(List list, {bool withReferenceable: false, Type type, bool useTypeInfo, bool withTypeInfo}) {
    return list
        .map((elem) => _encodeValue(elem,
            withReferenceable: withReferenceable, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo))
        .toList(growable: false);
  }

  String _encode(Object obj, {bool withReferenceable: false, bool useTypeInfo, bool withTypeInfo: false}) {
    Object val = toPrimaryObject(obj, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    return _codec.encode(val);
  }
}

==> serializer-0.8.2.tar.gz: lib/src/core/api.dart <==
// Copyright (c) 2016, the Serializer project authors.  Please see the AUTHORS file

// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'dart:convert';

import '../codecs/type_codec.dart';

bool isPrimaryType(Type obj) => obj == num || obj == String || obj == bool || obj == int || obj == double;

bool isPrimaryTypeString(String obj) =>
    obj == "num" || obj == "String" || obj == "bool" || obj == "int" || obj == "double";

final String MapTypeString = {}.runtimeType.toString();
final String ListTypeString = [].runtimeType.toString();

Map cleanNullInMap(Map<String, dynamic> map) {
  Iterable<String> keys = new List.from(map.keys);
  keys.forEach((String k) {
    if (map[k] == null) {
      map.remove(k);
    }
  });
  return map;
}

abstract class _Serializer {
  /// Return typeInfoKey value.
  String get typeInfoKey;

  /// Return useTypeInfo value.
  bool get useTypeInfo;

  /// Return if typeInfoKey must be enabled
  bool enableTypeInfo(bool useTypeInfo, bool withTypeInfo) =>
      useTypeInfo != null ? useTypeInfo || (withTypeInfo ?? false) : this.useTypeInfo || (withTypeInfo ?? false);

  /// Registers a [typeCodec] for the specific [type].
  addTypeCodec(Type type, TypeCodec typeCodec);

  /// Registers a map of [typeCodec].
  addAllTypeCodecs(Map<String, TypeCodec> typesCodecs);

  /// Checks if a TypeCodec is registered for the [type].
  bool hasTypeCodec(Type type);

  /// Get the TypeCodec for the specific [type]
  TypeCodec typeCodec(Type type);

  /// Checks if a class is registered as a Serializable class.
  bool isSerializable(Type type);

  /// Convert the object to a Map
  Map<String, dynamic> toMap(Object input, {bool useTypeInfo, bool withTypeInfo});

  /// Convert to a Map or a List recursively
  Object toPrimaryObject(Object input, {bool useTypeInfo, bool withTypeInfo});

  /// Encode the object to serialized string
  String encode(Object input, {bool useTypeInfo, bool withTypeInfo});

  /// Decode the object from a seriablized string
  Object decode(dynamic encoded, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo});

  /// Convert a serialized object to map
  Object fromMap(Map<String, dynamic> map, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo});

  /// Convert a serialized object's [list] to a list of the given [type]
  List fromList(List<dynamic> list, {Type type, bool useTypeInfo, bool withTypeInfo});
}

/// Utility class to access to the serializer api
class Serializer extends _Serializer {
  ///////////////////
  // Public
  /////////////////////////////////////////////////////////////////////////////

  /// Dump serializable classes
  //static dumpSerializables() => printAndDumpSerializables();

  final Codec _codec;
  final String _typeInfoKey;
  final bool _useTypeInfo;
  final Map<String, TypeCodec> _typeCodecs = <String, TypeCodec>{};

  String get typeInfoKey => _typeInfoKey;
  bool get useTypeInfo => _useTypeInfo;

  Serializer(
      {Map<String, TypeCodec> typesCodecs, Codec codec: json, String typeInfoKey: "@type", bool useTypeInfo: false})
      : _codec = codec,
        _typeInfoKey = typeInfoKey,
        _useTypeInfo = useTypeInfo {
    addAllTypeCodecs(typesCodecs);
  }

  factory Serializer.json({Map<String, TypeCodec> typesCodecs}) {
    return new Serializer(typesCodecs: typesCodecs, codec: json);
  }

  factory Serializer.typedJson({Map<String, TypeCodec> typesCodecs}) {
    return new Serializer(typesCodecs: typesCodecs, codec: json, useTypeInfo: true);
  }

  addTypeCodec(Type type, TypeCodec typeCodec) => _typeCodecs[type.toString()] = typeCodec;

  @override
  addAllTypeCodecs(Map<String, TypeCodec> typesCodecs) {
    _typeCodecs.addAll(typesCodecs ?? {});
  }

  bool hasTypeCodec(Type type) => _typeCodecs.values?.any((TypeCodec tc) => tc.type == type);

  TypeCodec typeCodec(Type type) =>
      _typeCodecs.values.firstWhere((TypeCodec tc) => tc.type == type, orElse: () => null);

  TypeCodec typeCodecFromString(String type) =>
      _typeCodecs.values.firstWhere((TypeCodec tc) => tc.typeInfo == type, orElse: () => null);

  bool isSerializable(Type type) => hasTypeCodec(type);

  Map<String, dynamic> toMap(Object input, {bool useTypeInfo, bool withTypeInfo}) =>
      toPrimaryObject(input, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo) as Map<String, dynamic>;

  /// Convert to a Map or a List recursively
  Object toPrimaryObject(Object input, {bool useTypeInfo, bool withTypeInfo}) =>
      _encodeValue(input, withReferenceable: true, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);

  /// Encode the object to serialized string
  String encode(Object input, {bool useTypeInfo, bool withTypeInfo}) =>
      _encode(input, withReferenceable: true, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);

  /// Decode the object from a seriablized string
  Object decode(dynamic encoded, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo}) =>
      _decode(encoded, type: type, mapOf: mapOf, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);

  /// Convert a serialized object to map
  Object fromMap(Map map, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo}) =>
      _fromMap(map, type: type, mapOf: mapOf, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);

  /// Convert a serialized object's [list] to a list of the given [type]
  List<dynamic> fromList(List list, {Type type, bool useTypeInfo, bool withTypeInfo}) =>
      _fromList(list, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);

  ///////////////////
  // Private
  /////////////////////////////////////////////////////////////////////////////
  List<Type> _findGenericOfMap(Type type) {
    String str = type.toString();
    RegExp reg = new RegExp(r"^Map<(.*)\ *,\ *(.*)>$");
    Iterable<Match> matches = reg.allMatches(str);
    if (matches == null || matches.isEmpty) {
      return null;
    }
    return [_decodeType(matches.first.group(1)), _decodeType(matches.first.group(2))];
  }

  Type _findGenericOfList(Type type) {
    String str = type.toString();
    RegExp reg = new RegExp(r"^List<(.*)>$");
    Iterable<Match> matches = reg.allMatches(str);
    if (matches == null || matches.isEmpty) {
      return null;
    }
    return _decodeType(matches.first.group(1));
  }

  String _getCorrectType(String name) {
    /*   String type = correspondingMinifiedTypes[name]; //todo: check minified
    if (type != null && _classes.containsKey(type)) {
      return type;
    }*/
    return name;
  }

  Type _decodeType(String name) {
    name = _getCorrectType(name);
    switch (name) {
      case "num":
        return num;
      case "String":
        return String;
      case "bool":
        return bool;
      case "int":
        return int;
      case "double":
        return double;
      case "Map":
      case "_JsonMap":
        return Map;
      case "List":
        return List;
      default:
        if (name == MapTypeString) {
          return Map;
        } else if (name == ListTypeString) {
          return List;
        }
        return typeCodecFromString(name)?.type;
    }
  }

  Object _decode(dynamic encoded, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo}) {
    if (encoded == null) {
      return null;
    }
    dynamic value = encoded;
    if (value is String) {
      var string = value as String;
      if (string.isEmpty) {
        return null;
      }
      if ((string.startsWith("{") && string.endsWith("}")) || (string.startsWith("[") && string.endsWith("]"))) {
        value = _codec.decode(encoded);
      } else {
        value = num.tryParse(encoded as String) ?? value;
      }
    }
    if (value is Map) {
      return _fromMap(value, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo, mapOf: mapOf);
    } else if (value is List) {
      return _fromList(value, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    }
    return _decodeValue(value, type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
  }

  Object _fromMap(Map map, {Type type, List<Type> mapOf, bool useTypeInfo, bool withTypeInfo}) {
    if (map == null || map.isEmpty) {
      return null;
    }

    if ((enableTypeInfo(useTypeInfo, withTypeInfo) || type == dynamic) && map.containsKey(_typeInfoKey)) {
      type = _decodeType(map[_typeInfoKey] as String);
    } else {
      type ??= Map;
    }

    // Only Map
    if (type == Map) {
      Type embedType = mapOf != null && mapOf.length >= 2 ? mapOf[1] : null;
      Map data = new Map();
      map.forEach((key, value) => data[key.toString()] =
          embedType == null && useTypeInfo == null ? value : _decodeValue(value, embedType, useTypeInfo: useTypeInfo));
      return data;
    }
    return _decodeValue(map, type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
  }

  List<dynamic> _fromList(List list, {Type type, bool useTypeInfo, bool withTypeInfo}) {
    List<dynamic> data = new List<dynamic>();
    list.forEach((value) => data.add(_decodeValue(value, type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo)));
    return data;
  }

  Object _decodeValue(Object value, Type type, {bool useTypeInfo, bool withTypeInfo}) {
    if (value == null) {
      return null;
    }
    String t = value.runtimeType.toString();
    if (value is Map && (_useTypeInfo == true || useTypeInfo == true) && value[typeInfoKey] != null) {
      t = value[typeInfoKey] as String;
    }
    type ??= _decodeType(t);
    if (isPrimaryType(type) ||
        value.runtimeType == type ||
        (isPrimaryType(value.runtimeType) && hasTypeCodec(type) == false)) {
      return value;
    } else if (hasTypeCodec(type) == true) {
      TypeCodec codec = typeCodec(type);
      dynamic val = codec.decode(value, serializer: this);
      return val;
    } else if (type.toString().startsWith("Map") || type.toString().startsWith("List")) {
      if (value is Map) {
        if (type.toString().startsWith("Map")) {
          return _fromMap(value,
              type: Map, mapOf: _findGenericOfMap(type), useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
        }
        return _fromMap(value, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
      } else if (value is List) {
        if (type.toString().startsWith("List") && value is List) {
          return _fromList(value, type: _findGenericOfList(type), useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
        }
        return _fromList(value, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
      }
    }

    return null;
  }

  Object _encodeValue(Object value, {bool withReferenceable: false, Type type, bool useTypeInfo, bool withTypeInfo}) {
    if (isPrimaryType(value.runtimeType) == true) {
      return value;
    } else if (hasTypeCodec(value.runtimeType) == true) {
      TypeCodec codec = typeCodec(value.runtimeType);
      return codec.encode(value, serializer: this, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    } else if (value is Map) {
      return _encodeMap(value,
          withReferenceable: withReferenceable, type: type, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    } else if (value is List) {
      return _encodeList(value,
          withReferenceable: withReferenceable, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    }
    return null;
  }

  Object _encodeMap(Map value, {bool withReferenceable: false, Type type, bool useTypeInfo, bool withTypeInfo}) {
    Map<String, dynamic> map = new Map<String, dynamic>();

    value.forEach((dynamic key, dynamic val) {
      map[key.toString()] = toPrimaryObject(val, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    });

    return map;
  }

  List _encodeList(List list, {bool withReferenceable: false, Type type, bool useTypeInfo, bool withTypeInfo}) {
    return list
        .map((elem) => _encodeValue(elem,
            withReferenceable: withReferenceable, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo))
        .toList(growable: false);
  }

  String _encode(Object obj, {bool withReferenceable: false, bool useTypeInfo, bool withTypeInfo: false}) {
    Object val = toPrimaryObject(obj, useTypeInfo: useTypeInfo, withTypeInfo: withTypeInfo);
    return _codec.encode(val) as String;
  }
}
